/**
 *   General USB communication stuff and EP0 control
 */

#include <fx2.h>
#include "irq.h"
#include "types.h"
#include "descriptors.h"
#include "delay.h"
#include "usb_ep0.h"


/* USB suspend request */
static volatile bit suspend = 0;
static volatile bit setup_data_available = 0;
static volatile bit enable_remote_wakeup = 0;

static xdata u8 * current_config_desc = usb_desc_config_full_speed;
static xdata u8 * other_speed_config = usb_desc_config_high_speed;

static u8 alternate_setting = 0;
static u8 actual_configuration = 0;



static
xdata u8* usb_get_string_desc(u8 idx)
{
	xdata u8 *s;

	switch (idx) {
	case 0:
		s = usb_desc_string_language;
		break;
	case 1:
		s = usb_desc_string_manufacturer;
		break;
	case 2:
		s = usb_desc_string_product;
		break;
	default:
		s = 0;
	}

	return s;
}


/* get address of EP control&status register */
static volatile
xdata u8 *usb_epcs(u8 ep)
{
	if (ep == 0x01)
		return &EP1OUTCS;
	else if (ep == 0x81)
		return &EP1INCS;

	ep &= 0x0e;
	ep /= 2;

	return &EP1INCS + ep;
}



static
void usb_disconnect_and_renumerate (void)
{
	USBCS |= 0x08;	/* disconnect */

	mdelay(1500);

	/* clear all pending irqs -- they're invalid now */
	USBIRQ = 0xff;
	EPIRQ = 0xff;
	EXIF &= ~0x10;

	USBCS |= 0x02;		/* renum */
	USBCS &= ~0x08;		/* reconnect */
}


void usb_ep0_init (void)
{
	current_config_desc[1] = USB_DT_CONFIG;
	other_speed_config[1] = USB_DT_OTHER_SPEED_CONFIG;

	/* ignore NACK interrupts for now, maybe we can use
	 * them later to save USB bandwidth...
	 */
	NAKIE = 0x00;           /* disable ping-nack irqs */
	IBNIE = 0x00;           /* disable in-bulk-nack irqs */

	syncdelay();
	EP0BCH = 0;
	EP0BCL = 0;
	syncdelay();

	EUSB = 1;               /* enable USB interrupts */
	USBIE |= 0x39;		/* HSGRANT URES SUDAV SUSP */

	INTSETUP |= 0x08;	/* enable irq2 autovectoring */

	EA = 1;

	/* renumerate if necessary */
	if (!(USBCS & 0x02))
		usb_disconnect_and_renumerate();

	/* reconnect */
	USBCS &= ~0x08;
}


#pragma SAVE
#pragma NOIV

/**
 *  process setup data + setup requests
 */
void usb_ep0_process_request (void)
{
	xdata u8 *desc;

	if (!setup_data_available)
		return;

	setup_data_available = 0;

	switch (SETUPDAT[1]) {
	case USB_REQ_GET_DESCRIPTOR:
		switch (SETUPDAT[3]) {
		case USB_DT_DEVICE:
			SUDPTRH = msb(usb_desc_device);
			SUDPTRL = lsb(usb_desc_device);
			break;
		case USB_DT_DEVICE_QUALIFIER:
			SUDPTRH = msb(usb_desc_device_qualifier);
			SUDPTRL = lsb(usb_desc_device_qualifier);
			break;
		case USB_DT_CONFIG:
			SUDPTRH = msb(current_config_desc);
			SUDPTRL = lsb(current_config_desc);
			break;
		case USB_DT_OTHER_SPEED_CONFIG:
			SUDPTRH = msb(other_speed_config);
			SUDPTRL = lsb(other_speed_config);
			break;
		case USB_DT_STRING:
			desc = usb_get_string_desc(SETUPDAT[2]);
			if (desc) {
				SUDPTRH = msb(desc);
				SUDPTRL = lsb(desc);
			} else
				EP0CS |= 0x01;	/* stall */
			break;
		default:
			/* invalid request */
			EP0CS |= 0x01;	/* stall */
		}
		break;

	case USB_REQ_GET_INTERFACE:
		EP0BUF[0] = alternate_setting;
		syncdelay();
		EP0BCH = 0;
		EP0BCL = 1;
		syncdelay();
		break;

	case USB_REQ_SET_INTERFACE:
		if (SETUPDAT[2] < 2) {
//			alternate_setting = SETUPDAT[2];
//			switch (alternate_setting) {
//			case 0:
//				usb_ep2468_gpif_init();
//				break;
//			case 1:
//				usb_ep2hs_gpif_init();
//				break;
//			default:
//				;
//			}
		} else
			EP0CS |= 0x01;	/* stall */
		break;

	case USB_REQ_SET_CONFIGURATION:
		if (SETUPDAT[2] > 1)
			EP0CS |= 0x01;	/* stall */ 
		else
			actual_configuration = SETUPDAT[2];
		break;

	case USB_REQ_GET_CONFIGURATION:
		EP0BUF[0] = actual_configuration;
		syncdelay();
		EP0BCH = 0;
		EP0BCL = 1;
		syncdelay();
		break;

	case USB_REQ_GET_STATUS:
		switch (SETUPDAT[0]) {
		case USB_DT_DEVICE:
			EP0BUF[0] = (enable_remote_wakeup << 1) | 0x01 /* self powered */;
			EP0BUF[1] = 0;
			syncdelay();
			EP0BCH = 0;
			EP0BCL = 2;
			syncdelay();
			break;
		case USB_DT_INTERFACE:
			EP0BUF[0] = 0;
			EP0BUF[1] = 0;
			syncdelay();
			EP0BCH = 0;
			EP0BCL = 2;
			syncdelay();
			break;
		case USB_DT_ENDPOINT:
			EP0BUF[0] = (*usb_epcs(SETUPDAT[4])) & 0x01; /* stalled? */
			EP0BUF[1] = 0;
			syncdelay();
			EP0BCH = 0;
			EP0BCL = 2;
			syncdelay();
			break;
		default:
			/* invalid command */
			EP0CS |= 0x01;	/* stall */
		}
		break;

	case USB_REQ_CLEAR_FEATURE:
		switch (SETUPDAT[0]) {
		case USB_FT_DEVICE_REMOTE_WAKEUP:
			if (SETUPDAT[2] == 1)
				enable_remote_wakeup = FALSE;
			else
				EP0CS |= 0x01;	/* stall */
			break;
		case USB_FT_ENDPOINT_HALT:
			if (SETUPDAT[2] == 0) {
				u8 ep;
				ep = SETUPDAT[4];
				*usb_epcs(ep) &= ~0x01;
				/* reset data toggle */
				TOGCTL = (((ep & 0x80) >> 3) + (ep & 0x0f));
				TOGCTL |= 0x20;
			} else
				EP0CS |= 0x01;	/* stall */
			break;
		default:
			;
		}
		break;

	case USB_REQ_SET_FEATURE:
		switch (SETUPDAT[0]) {
		case USB_FT_DEVICE_REMOTE_WAKEUP:
			if (SETUPDAT[2] == 1)
				enable_remote_wakeup = TRUE;
			else
				EP0CS |= 0x01;	/* stall */
			break;
		case USB_FT_ENDPOINT_HALT:
			*usb_epcs(SETUPDAT[4]) |= 0x01; /* stall */
			break;
		}
		break;

	//case 0xa2:
		//TODO: write ID EEPROM
		//really??

	/**
	 *  we define our own proprietary config request to allow memory dumps
	 *  for the high memory area. 0xa3 is within the Cypress reserved range
	 *  0xa1...0xaf (see section 3.7 in the Technical Reference Manual)
	 */
	case 0xa3:
	{
		u8 i, len = SETUPDAT[6];

		AUTOPTRL1 = SETUPDAT[2];
		AUTOPTRH1 = SETUPDAT[3];
		for (i=0; i<len; i++)
			EP0BUF[i] = XAUTODAT1;
		syncdelay();
		EP0BCH = 0;
		EP0BCL = len;
		syncdelay();
		break;
	}
	
	case 0xa4: {
		EP0BUF[0] = EP6CS;
		EP0BUF[1] = EP6FIFOFLGS;
		EP0BUF[2] = EP6FIFOBCL;
		EP0BUF[3] = EP6FIFOBCH;
		
		syncdelay();
		
		EP0BCH = 0;
		EP0BCL = 4;
		
		syncdelay();
		
		break;
	}

	default:
		/* invalid command */
		EP0CS |= 0x01;	/* stall */
	}

	/* clear hand shake NAK (-> ACK) */
	EP0CS |= 0x80;
}


void usb_sudav_irq (void) interrupt 8 using 1
{
	setup_data_available = 1;
	/* ack irq */
	EXIF &= ~0x10;
	USBIRQ = 0x01;
}


void usb_suspend_irq (void) interrupt 8 using 1
{
	suspend = TRUE;
	/* ack irq */
	EXIF &= ~0x10;
	USBIRQ = 0x08;
}


void usb_usbreset_irq (void) interrupt 8 using 1
{
	if (USBCS & 0x80 /* high speed */) {
		current_config_desc = usb_desc_config_high_speed;
		other_speed_config = usb_desc_config_full_speed;
	} else {
		current_config_desc = usb_desc_config_full_speed;
		other_speed_config = usb_desc_config_high_speed;
	}

	current_config_desc[1] = USB_DT_CONFIG;
	other_speed_config[1] = USB_DT_OTHER_SPEED_CONFIG;

	/* ack irq */
	EXIF &= ~0x10;
	USBIRQ = 0x10;
}


/**
 *  high speed access granted
 */
void usb_hsgrant_irq (void) interrupt 8 using 1
{
	if (USBCS & 0x80 /* high speed */) {
		current_config_desc = usb_desc_config_high_speed;
		other_speed_config = usb_desc_config_full_speed;
	} else {
		current_config_desc = usb_desc_config_full_speed;
		other_speed_config = usb_desc_config_high_speed;
	}

	current_config_desc[1] = USB_DT_CONFIG;
	other_speed_config[1] = USB_DT_OTHER_SPEED_CONFIG;

	/* ack irq */
	EXIF &= ~0x10;
	USBIRQ = 0x20;
}

#pragma RESTORE



static
void usb_suspend (void)
{
	WAKEUPCS |= 0xc0;	/* clear wakeup condition */
	SUSPEND = 0xff;		/* do suspend */
	PCON |= 0x01;		/* idle, stop the FX2's CPU clock */

	_asm
	nop
	nop
	nop
	nop
	nop
	_endasm;
}


#define usb_wakeup() \
	(((WAKEUPCS & 0x01) && (WAKEUPCS & 0x40)) || \
	 ((WAKEUPCS & 0x02) && (WAKEUPCS & 0x80)))


void usb_resume (void)
{
	if (usb_wakeup()) {
		USBCS |= 0x01;	/* do resume */
		mdelay(20);
		USBCS &= ~0x01;
	}
}


void usb_check_suspend_state (void)
{
	if (suspend) {
		suspend = 0;
		
		do {
			usb_suspend();
		} while (!enable_remote_wakeup && usb_wakeup());

		usb_resume();
	}
}
